#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Deprecated trio plugin."""

from nuitka.plugins.PluginBase import NuitkaPluginBase


class NuitkaPluginTrio(NuitkaPluginBase):
    plugin_name = "trio"
    plugin_desc = "Deprecated, was once required by 'trio' package."
    plugin_category = "package-support,obsolete"

    @classmethod
    def isDeprecated(cls):
        return True


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
